@@ -142,7 +142,7 @@ sub get_parser ( $self, $system = undef, $nofallback = undef ) {
     }
     elsif ( $system ne $::config->{notes}->{system} ) {
 	my $p = ChordPro::Chords::Parser::Common->new
-	  ( { notes => $system } );
+	  ( { notes => { system => $system } } );
 	return $parsers{$system} = $p;
     }
     elsif ( $system ) {
@@ -211,7 +211,7 @@ sub parse_chord ( $self, $chord ) {
 		 name   => $chord };
 
     my $bass = "";
-    if ( $chord =~ m;^(.*)/(.*); ) {
+    if ( $chord =~ m;^(.*)/($self->{n_pat})$; ) {
 	$chord = $1;
 	$bass = $2;
     }
@@ -543,6 +543,7 @@ sub new {
     my $self = bless { chord_cache => {} } => $pkg;
     $self->{system} = "nashville";
     $self->{target} = 'ChordPro::Chord::Nashville';
+    $self->{intervals} = 12;
     warn("Chords: Created parser for ", $self->{system}, "\n")
       if $::options->{verbose} && $::options->{verbose} > 1;
     return $parsers{$self->{system}} = $self;
@@ -655,6 +656,7 @@ sub new {
     my $self = bless { chord_cache => {} } => $pkg;
     $self->{system} = "roman";
     $self->{target} = 'ChordPro::Chord::Roman';
+    $self->{intervals} = 12;
     warn("Chords: Created parser for ", $self->{system}, "\n")
       if $::options->{verbose} && $::options->{verbose} > 1;
     return $parsers{$self->{system}} = $self;
@@ -838,10 +840,32 @@ sub kbkeys {
     $_[0]->{keys} = ChordPro::Chords::get_keys($_[0]);
 }
 
+sub chord_display ( $self, $default ) {
+
+    use String::Interpolate::Named;
+
+    my $res = $self->name;
+    my $args = {};
+    $self->flat_copy( $args, $self->{display} // $self );
+
+    for my $fmt ( $default,
+		  $self->{format},
+		  $self->{chordformat} ) {
+	next unless $fmt;
+	$args->{root} = lc($args->{root}) if $self->is_note;
+	$args->{formatted} = $res;
+	$res = interpolate( { args => $args }, $fmt );
+    }
+
+    # Substitute musical symbols if wanted.
+    return $::config->{settings}->{truesf} ? $self->fix_musicsyms($res) : $res;
+}
+
 sub flat_copy ( $self, $ret, $o, $pfx = "" ) {
     while ( my ( $k, $v ) = each %$o ) {
-	if ( $k eq "orig" ) {
+	if ( $k eq "orig" || $k eq "xc" || $k eq "xp" ) {
 	    $self->flat_copy( $ret, $v, "$k.$pfx");
+	    $ret->{"$k.${pfx}formatted"} = $v->chord_display;
 	}
 	else {
 	    $ret->{"$pfx$k"} = $v;
@@ -901,8 +925,6 @@ package ChordPro::Chord::Common;
 
 our @ISA = qw( ChordPro::Chord::Base );
 
-use String::Interpolate::Named;
-
 # Show reconstructs the chord from its constituents.
 # Result is canonical.
 sub show {
@@ -983,18 +1005,24 @@ sub transcode ( $self, $xcode, $key_ord = 0 ) {
     my $p = $self->{parser}->get_parser($xcode);
     die("OOPS ", $p->{system}, " $xcode") unless $p->{system} eq $xcode;
     $info->{parser} = $p;
-    $info->{root_ord} -= $key_ord if $key_ord && $p->movable;
+    if ( $key_ord && $p->movable ) {
+	$info->{root_ord} -= $key_ord % $p->intervals;
+    }
 #    $info->{$_} = $p->{$_} for qw( ns_tbl nf_tbl ns_canon nf_canon );
-    $info->{root_canon} = $info->{root} =
-      $p->root_canon( $info->{root_ord},
-		      $info->{root_mod} >= 0,
-		      $info->{qual_canon} eq "-" );
+    unless ( $self->{rootless} ) {
+	$info->{root_canon} = $info->{root} =
+	  $p->root_canon( $info->{root_ord},
+			  $info->{root_mod} >= 0,
+			  $info->{qual_canon} eq "-" );
+    }
     if ( $p->{system} eq "roman" && $info->{qual_canon} eq "-" ) {
 	# Minor quality is in the root name.
 	$info->{qual_canon} = $info->{qual} = "";
     }
     if ( $self->{bass} && $self->{bass} ne "" ) {
-	$info->{bass_ord} -= $key_ord if $key_ord && $p->movable;
+	if ( $key_ord && $p->movable ) {
+	    $info->{bass_ord} -= $key_ord % $p->intervals;
+	}
 	$info->{bass_canon} = $info->{bass} =
 	  $p->root_canon( $info->{bass_ord}, $info->{bass_mod} >= 0 );
     }
@@ -1008,23 +1036,10 @@ sub transcode ( $self, $xcode, $key_ord = 0 ) {
 
 sub chord_display ( $self ) {
 
-    # $self->dump;
-
-    my $res = $self->name;
-    my $args = {};
-    $self->flat_copy( $args, $self->{display} // $self );
-
-    for my $fmt ( $::config->{settings}->{"chord-format"},
-		  $self->{format},
-		  $self->{chordformat} ) {
-	next unless $fmt;
-	$args->{root} = lc($args->{root}) if $self->is_note;
-	$args->{formatted} = $res;
-	$res = interpolate( { args => $args }, $fmt );
-    }
-
-    # Substitute musical symbols if wanted.
-    return $::config->{settings}->{truesf} ? $self->fix_musicsyms($res) : $res;
+    $self->SUPER::chord_display
+      ( $::config->{"chord-formats"}->{common}
+	// $::config->{settings}->{"chord-format"}
+	// "%{name}" );
 }
 
 ################ Chord objects: Nashville ################
@@ -1033,9 +1048,7 @@ package ChordPro::Chord::Nashville;
 
 our @ISA = 'ChordPro::Chord::Base';
 
-use String::Interpolate::Named;
-
-sub transpose ( $self ) { $self }
+sub transpose ( $self, $dummy1, $dummy2=0 ) { $self }
 
 sub show {
     Carp::croak("call canonical instead of show");
@@ -1050,22 +1063,10 @@ sub canonical ( $self ) {
 }
 
 sub chord_display ( $self ) {
-    if ( $self->{format} ) {
-	####TODO
-	my $fmt = $self->{format} || $::config->{settings}->{"chord-format"};
-	if ( $fmt ) {
-	    my $args = {};
-	    $self->flat_copy( $args, $self );
-	    return interpolate( { args => $args }, $fmt );
-	}
-    }
 
-    my $res = $self->{root_canon} .
-      "<sup>" . $self->{qual} . $self->{ext} . "</sup>";
-    if ( $self->{bass} && $self->{bass} ne "" ) {
-	$res .= "<sub>/" . lc($self->{bass}) . "</sub>";
-    }
-    return $res;
+    $self->SUPER::chord_display
+      ( $::config->{"chord-formats"}->{nashville}
+	// "%{name}" );
 }
 
 ################ Chord objects: Roman ################
@@ -1074,9 +1075,7 @@ package ChordPro::Chord::Roman;
 
 our @ISA = 'ChordPro::Chord::Base';
 
-use String::Interpolate::Named;
-
-sub transpose ( $self ) { $self }
+sub transpose ( $self, $dummy1, $dummy2=0 ) { $self }
 
 sub show {
     Carp::croak("call canonical instead of show");
@@ -1091,24 +1090,10 @@ sub canonical ( $self ) {
 }
 
 sub chord_display ( $self ) {
-    if ( $self->{format} ) {
-	####TODO
-	my $fmt = $self->{format} || $::config->{settings}->{"chord-format"};
-	if ( $fmt ) {
-	    my $args = {};
-	    $self->flat_copy( $args, $self );
-	    return interpolate( { args => $args }, $fmt );
-	}
-	return $self->canonical;
-    }
 
-    my $res = $self->{root_canon};
-    $res .= "<sup>" . $self->{qual} . $self->{ext} . "</sup>"
-      if $self->{qual};
-    if ( $self->{bass} && $self->{bass} ne "" ) {
-	$res .= "<sub>/" . lc($self->{bass}) . "</sub>";
-    }
-    return $res;
+    $self->SUPER::chord_display
+      ( $::config->{"chord-formats"}->{roman}
+	// "%{name}" );
 }
 
 ################ Chord objects: Annotations ################
@@ -1119,8 +1104,8 @@ use String::Interpolate::Named;
 
 our @ISA = 'ChordPro::Chord::Base';
 
-sub transpose ( $self ) { $self }
-sub transcode ( $self ) { $self }
+sub transpose ( $self, $dummy1, $dummy2=0 ) { $self }
+sub transcode ( $self, $dummy1, $dummy2=0 ) { $self }
 
 sub canonical ( $self ) {
     my $res = $self->{text};
