@@ -1,6 +1,6 @@
 // break.js - module to handle the %%break command
 //
-// Copyright (C) 2018-2023 Jean-Francois Moine
+// Copyright (C) 2018-2024 Jean-Francois Moine
 //
 // This file is part of abc2svg.
 //
@@ -30,7 +30,7 @@ abc2svg.break = {
 	// get the %%break parameters
 	get_break: function(parm) {
 	    var	C = abc2svg.C,
-		b, c, d, sq,
+		b, c, d, n, sq,
 		a = parm.split(/[ ,]/),
 		cfmt = this.cfmt()
 
@@ -66,8 +66,8 @@ abc2svg.break = {
 
 	// insert the EOLs of %%break
 	do_break: function() {
-	    var	i, m, t, brk, seq,
-		voice_tb = this.get_voice_tb()
+	    var	i, m, t, brk, s, seq,
+		voice_tb = this.get_voice_tb(),
 		v = this.get_cur_sy().top_voice,
 		s1 = voice_tb[v].sym,
 		cfmt = this.cfmt()
